# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import logging

from django.contrib.auth.models import User
from django.db import models


class BugzillaUserMap(models.Model):
    user = models.OneToOneField(User)
    bugzilla_user_id = models.IntegerField(unique=True, db_index=True)


def get_or_create_bugzilla_users(user_data):
    # All users will have a stored password of "!", which Django uses to
    # indicate an invalid password.
    users = []

    for user in user_data['users']:
        bz_user_id = user['id']
        email = user['email']
        real_name = user['real_name']
        can_login = user['can_login']

        try:
            bugzilla_user_map = BugzillaUserMap.objects.get(
                bugzilla_user_id=bz_user_id)
        except BugzillaUserMap.DoesNotExist:
            user = User(username=email, password='!', first_name=real_name,
                        email=email, is_active=can_login)
            user.save()
            bugzilla_user_map = BugzillaUserMap(user=user,
                                                bugzilla_user_id=bz_user_id)
            bugzilla_user_map.save()
        else:
            modified = False
            user = bugzilla_user_map.user

            if user.username != email:
                user.username = email
                user.email = email
                modified = True

            if user.first_name != real_name:
                user.first_name = real_name
                modified = True

            if user.is_active != can_login:
                user.is_active = can_login
                modified = True

            if modified:
                user.save()

        users.append(user)
    return users


from reviewboard.reviews.errors import PublishError

#
# Review request errors.
#
class InvalidBugsError(PublishError):
    def __init__(self):
        PublishError.__init__(self, 'Exactly one bug ID must be provided.')


class InvalidBugIdError(PublishError):
    def __init__(self, bug_id):
        PublishError.__init__(self, 'Invalid bug ID "%s".' % bug_id)


class ConfidentialBugError(PublishError):
    def __init__(self):
        PublishError.__init__(self, 'This bug is confidential; please attach '
                              'the patch directly to the bug.')


#
# Bugzilla errors.
#
class BugzillaError(Exception):
    def __init__(self, msg, fault_code=None):
        super(BugzillaError, self).__init__(msg)
        self.msg = msg
        self.fault_code = fault_code


class BugzillaUrlError(BugzillaError):
    def __init__(self):
        BugzillaError.__init__(self, 'No Bugzilla URL provided in rbbz '
                               'configuration.')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from django import forms
from djblets.siteconfig.forms import SiteSettingsForm


class BugzillaAuthSettingsForm(SiteSettingsForm):

    auth_bz_xmlrpc_url = forms.CharField(
        label="Bugzilla XMLRPC URL",
        help_text="URL for your Bugzilla installation's XMLRPC interface",
        required=True)

    class Meta:
        title = "Bugzilla Backend Settings"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import xmlrpclib

from djblets.siteconfig.models import SiteConfiguration
from djblets.util.decorators import simple_decorator

from rbbz.errors import BugzillaError, BugzillaUrlError
from rbbz.transports import bugzilla_transport


@simple_decorator
def xmlrpc_to_bugzilla_errors(func):
    def _transform_errors(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except xmlrpclib.Fault as e:
            raise BugzillaError(e.faultString, e.faultCode)
    return _transform_errors


class Bugzilla(object):
    """
    Interface to a Bugzilla system.

    At the moment this uses the XMLRPC API.  It should probably be converted
    to REST at some point.

    General FIXME: try to get more specific errors out of xmlrpclib.Fault
    exceptions.
    """

    user_fields = ['id', 'email', 'real_name', 'can_login']

    def __init__(self, login=None, logincookie=None, xmlrpc_url=None):
        if xmlrpc_url:
            self.xmlrpc_url = xmlrpc_url
        else:
            siteconfig = SiteConfiguration.objects.get_current()
            self.xmlrpc_url = siteconfig.get('auth_bz_xmlrpc_url')

        if not self.xmlrpc_url:
            raise BugzillaUrlError('no XMLRPC URL')

        self._transport = None
        self._proxy = None

        if logincookie:
            self.transport.set_bugzilla_cookies(login, logincookie)

    def cookies(self):
        return self.transport.bugzilla_cookies()

    @xmlrpc_to_bugzilla_errors
    def log_in(self, username, password, cookie=False):
        if cookie:
            # Username and password are actually bugzilla cookies.
            self.transport.set_bugzilla_cookies(username, password)
            user_id = username
        else:
            self.transport.remove_bugzilla_cookies()

            try:
                result = self.proxy.User.login({'login': username,
                                                'password': password})

            except xmlrpclib.Fault as e:
                if e.faultCode == 300 or e.faultCode == 301:
                    # Invalid username/password or account disabled
                    return None
                raise

            user_id = result['id']

        params = {'ids': [user_id], 'include_fields': self.user_fields,
                  'groups': ['reviewboard']}

        try:
            return self.proxy.User.get(params)
        except xmlrpclib.Fault as e:
            if e.faultCode == 804:
                # User isn't a member of the requisite group.
                return None
            raise

    @xmlrpc_to_bugzilla_errors
    def get_user(self, username):
        params = {'names': [username], 'include_fields': self.user_fields}
        return self.proxy.User.get(params)

    @xmlrpc_to_bugzilla_errors
    def query_users(self, query):
        params = {'match': [query], 'include_fields': self.user_fields}
        return self.proxy.User.get(params)

    @xmlrpc_to_bugzilla_errors
    def post_comment(self, bug_id, comment):
        params = {
            'id': bug_id,
            'comment': comment
        }
        return self.proxy.Bug.add_comment(params)

    @xmlrpc_to_bugzilla_errors
    def is_bug_confidential(self, bug_id):
        params = {'ids': [bug_id], 'include_fields': ['groups']}

        try:
            groups = self.proxy.Bug.get(params)['bugs'][0]['groups']
        except xmlrpclib.Fault as e:
            if e.faultCode == 102:
                return True
            raise

        return bool(groups)

    @xmlrpc_to_bugzilla_errors
    def post_rb_url(self, bug_id, review_id, summary, description, url,
                    reviewers):
        """Posts a new attachment containing a review-request URL, or updates
           an existing one."""

        reviewers = set(reviewers)
        params = {}
        flags = []
        rb_attachment = None
        attachments = self.get_rb_attachments(bug_id)

        # Find the associated attachment, then go through the review flags.
        # If a review flag status is '?' and the requestee is still in the
        # given reviewers, then update the flag.
        # If the review flag status is other than '?', or if it is a feedback
        # flag, clear it.

        for a in attachments:
            if a['data'] != url:
                continue

            rb_attachment = a

            for f in a.get('flags', []):
                if f['name'] != 'review' and f['name'] != 'feedback':
                    continue

                if (f['name'] == 'review' and 'requestee' in f
                    and f['requestee'] in reviewers):
                    flags.append({'id': f['id'], 'name': 'review',
                                  'status': '?',
                                  'requestee': f['requestee']})
                    reviewers.remove(f['requestee'])
                else:
                    flags.append({'id': f['id'], 'status': 'X'})

            break

        # Add flags for new reviewers.

        for r in reviewers:
            flags.append({'name': 'review', 'status': '?', 'requestee': r,
                          'new': True})

        if rb_attachment:
            params['ids'] = [rb_attachment['id']]
        else:
            params['ids'] = [bug_id]
            params['data'] = url
            params['content_type'] = 'text/plain'

        params['file_name'] = 'reviewboard-%d-url.txt' % review_id
        params['summary'] = summary
        params['comment'] = description
        if flags:
            params['flags'] = flags

        if rb_attachment:
            self.proxy.Bug.update_attachment(params)
        else:
            self.proxy.Bug.add_attachment(params)

    @xmlrpc_to_bugzilla_errors
    def get_rb_attachments(self, bug_id):
        """Get all attachments that contain review-request URLs."""

        params = {
            'ids': [bug_id],
            'include_fields': ['id', 'data', 'content_type', 'is_obsolete',
                               'flags']
        }

        return [a for a
                in self.proxy.Bug.attachments(params)['bugs'][str(bug_id)]
                if not a['is_obsolete'] and
                a['content_type'] == 'text/x-review-board-request']

    @xmlrpc_to_bugzilla_errors
    def r_plus_attachment(self, bug_id, reviewer, rb_url):
        """Set a review flag to "+"."""

        rb_attachment = None
        attachments = self.get_rb_attachments(bug_id)

        for a in attachments:
            if a.get('data') == rb_url:
                rb_attachment = a
                break

        if not rb_attachment:
            return

        flags = rb_attachment.get('flags', [])
        new_flag = {'name': 'review', 'status': '+'}

        for f in flags:
            if f['name'] == 'review' and f.get('requestee') == reviewer:
                new_flag['id'] = f['id']
                break
        else:
            new_flag['new'] = True
            new_flag['requestee'] = reviewer

        params = {
            'ids': [rb_attachment['id']],
            'flags': [new_flag]
        }

        self.proxy.Bug.update_attachment(params)

    @property
    def transport(self):
        if self._transport is None:
            self._transport = bugzilla_transport(self.xmlrpc_url)

        return self._transport

    @property
    def proxy(self):
        if self._proxy is None:
            self._proxy = xmlrpclib.ServerProxy(self.xmlrpc_url, self.transport)

        return self._proxy



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json

from django.contrib.sites.models import Site

from djblets.siteconfig.models import SiteConfiguration
from djblets.util.decorators import simple_decorator

from reviewboard.extensions.base import Extension
from reviewboard.extensions.hooks import AuthBackendHook, SignalHook
from reviewboard.reviews.errors import PermissionError, PublishError
from reviewboard.reviews.models import ReviewRequest
from reviewboard.reviews.signals import (reply_publishing,
                                         review_publishing,
                                         review_request_closed,
                                         review_request_publishing)
from reviewboard.site.urlresolvers import local_site_reverse

from rbbz.auth import BugzillaBackend
from rbbz.bugzilla import Bugzilla
from rbbz.diffs import build_plaintext_review
from rbbz.errors import (BugzillaError,
                         ConfidentialBugError,
                         InvalidBugIdError)
from rbbz.middleware import BugzillaCookieAuthMiddleware


BZIMPORT_PREFIX = "bz://"
AUTO_CLOSE_DESCRIPTION = """
Discarded automatically because parent review request was discarded.
"""


class BugzillaExtension(Extension):
    middleware = [BugzillaCookieAuthMiddleware]

    def initialize(self):
        AuthBackendHook(self, BugzillaBackend)
        SignalHook(self, review_request_publishing,
                   on_review_request_publishing)
        SignalHook(self, review_publishing, on_review_publishing)
        SignalHook(self, reply_publishing, on_reply_publishing)
        SignalHook(self, review_request_closed, on_review_request_closed)


def review_request_url(review_request, site=None, siteconfig=None):
    if not site:
        site = Site.objects.get_current()

    if not siteconfig:
        siteconfig = SiteConfiguration.objects.get_current()

    return '%s://%s%s%s' % (
        siteconfig.get('site_domain_method'), site.domain,
        local_site_reverse('root').rstrip('/'),
        review_request.get_absolute_url())


def is_review_request_pushed(review_request):
    return str(review_request.extra_data.get('p2rb', False)) == "True"


def is_review_request_squashed(review_request):
    squashed = str(review_request.extra_data.get('p2rb.is_squashed', False))
    return squashed == "True"


@simple_decorator
def bugzilla_to_publish_errors(func):
    def _transform_errors(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BugzillaError as e:
            raise PublishError('Bugzilla error: %s' % e.msg)
    return _transform_errors


@bugzilla_to_publish_errors
def on_review_request_publishing(user, review_request_draft, **kwargs):
    review_request = review_request_draft.get_review_request()

    # skip review requests that were not pushed
    if not is_review_request_pushed(review_request):
        return

    # The reviewid passed through p2rb is, for Mozilla's instance anyway, also
    # the bug ID.
    bug_id = review_request_draft.extra_data.get('p2rb.identifier', None)

    if bug_id.startswith(BZIMPORT_PREFIX):
        bug_id = bug_id[len(BZIMPORT_PREFIX):]

    try:
        bug_id = int(bug_id)
    except (TypeError, ValueError):
        raise InvalidBugIdError(bug_id)

    b = Bugzilla(user.bzlogin, user.bzcookie)

    try:
        if b.is_bug_confidential(bug_id):
            raise ConfidentialBugError
    except BugzillaError as e:
        # Special cases:
        #   100: Invalid Bug Alias
        #   101: Bug does not exist
        if e.fault_code and (e.fault_code == 100 or e.fault_code == 101):
            raise InvalidBugIdError(bug_id)
        raise

    # At this point, we know that the bug ID that we've got
    # is valid and accessible.
    review_request_draft.bugs_closed = str(bug_id)

    reviewers = [x.get_username() for x in
                 review_request_draft.target_people.all()]

    # Don't make attachments for child review requests, otherwise,
    # Bugzilla gets inundatated with lots of patches, and the squashed
    # one is the only one we want to post there.
    if is_review_request_squashed(review_request):
        b.post_rb_url(bug_id,
                      review_request.id,
                      review_request_draft.summary,
                      review_request_draft.description,
                      review_request_url(review_request),
                      reviewers)
        # Publish and child review requests that are either not
        # public, or have drafts.
        for child in generate_child_review_requests(review_request):
            try:
                if child.status == ReviewRequest.DISCARDED:
                    child.reopen(user=user)
                if child.get_draft(user=user) or not child.public:
                    child.publish(user=user)
            except PermissionError as e:
                logging.error('Could not reopen or publish child review '
                              'request with id %s because of error %s'
                              % (child.id, e))


def on_review_publishing(user, review, **kwargs):
    review_request = review.review_request

    # skip review requests that were not pushed
    if not is_review_request_pushed(review_request):
        return

    bug_id = int(review_request.get_bug_list()[0])
    site = Site.objects.get_current()
    siteconfig = SiteConfiguration.objects.get_current()
    b = Bugzilla(user.bzlogin, user.bzcookie)

    b.post_comment(bug_id, build_plaintext_review(review, {"user": user}))

    if review.ship_it and is_review_request_squashed(review_request):
        b.r_plus_attachment(bug_id, review.user.username,
                            review_request_url(review_request, site,
                                               siteconfig))


def on_reply_publishing(user, reply, **kwargs):
    review_request = reply.review_request

    # skip review requests that were not pushed
    if not is_review_request_pushed(review_request):
        return

    bug_id = int(review_request.get_bug_list()[0])
    b = Bugzilla(user.bzlogin, user.bzcookie)

    b.post_comment(bug_id, build_plaintext_review(reply, {"user": user}))


def on_review_request_closed(user, review_request, type, **kwargs):
    if (is_review_request_squashed(review_request) and
            type == ReviewRequest.DISCARDED):
        # At the point of discarding, it's possible that if this review
        # request was never published, that most of the fields are empty
        # (See https://code.google.com/p/reviewboard/issues/detail?id=3465).
        # Luckily, the extra_data is still around, and more luckily, it's
        # not exposed in the UI for user-meddling. We can find all of the
        # child review requests via extra_data.p2rb.commits.
        for child in generate_child_review_requests(review_request):
            child.close(ReviewRequest.DISCARDED,
                        user=user,
                        description=AUTO_CLOSE_DESCRIPTION)

        # Next, we clear out the commit_id on the squashed review request
        # so that someone can post follow-up work with the same review
        # identifier in the future. This value is, however, still being
        # stored in extra_data under "p2rb.identifier".
        review_request.commit = None
        review_request.save()


def generate_child_review_requests(squashed_review_request):
    if not is_review_request_squashed(squashed_review_request):
        return

    commits = squashed_review_request.extra_data['p2rb.commits']
    if commits:
        commits = json.loads(commits)
        for commit_tuple in commits:
            child_commit_id = commit_tuple[0]
            child_request_id = commit_tuple[1]
            try:
                child = ReviewRequest.objects.get(pk=child_request_id)
                yield child
            except ReviewRequest.DoesNotExist:
                logging.error('Could not retrieve child review request '
                              'with id %s belonging to commit %s because '
                              'it does not appear to exist.'
                              % (child_request_id, child_commit_id))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import urlparse
import xmlrpclib


class CookieTransportMixin:
    """A Transport request method that retains cookies over its lifetime.

    Adapted directly from http://www.lunch.org.uk/wiki/xmlrpccookies with
    permission from the author.

    The regular xmlrpclib transports ignore cookies, which causes
    a bit of a problem when you need a cookie-based login, as with
    the Bugzilla XMLRPC interface.

    So this is a helper for defining a Transport which looks for
    cookies being set in responses and saves them to add to all future
    requests.
    """

    # Inspiration drawn from
    # http://blog.godson.in/2010/09/how-to-make-python-xmlrpclib-client.html
    # http://www.itkovian.net/base/transport-class-for-pythons-xml-rpc-lib/
    #
    # Note this must be an old-style class so that __init__ handling works
    # correctly with the old-style Transport class. If you make this class
    # a new-style class, Transport.__init__() won't be called.

    cookies = []

    def send_cookies(self, connection):
        if self.cookies:
            for cookie in self.cookies:
                connection.putheader("Cookie", cookie)

    def request(self, host, handler, request_body, verbose=0):
        self.verbose = verbose

        # issue XML-RPC request
        h = self.make_connection(host)
        if verbose:
            h.set_debuglevel(1)

        self.send_request(h, handler, request_body)
        self.send_host(h, host)
        self.send_cookies(h)
        self.send_user_agent(h)
        self.send_content(h, request_body)

        # Deal with differences between Python 2.4-2.6 and 2.7.
        # In the former h is a HTTP(S). In the latter it's a
        # HTTP(S)Connection. Luckily, the 2.4-2.6 implementation of
        # HTTP(S) has an underlying HTTP(S)Connection, so extract
        # that and use it.
        try:
            response = h.getresponse()
        except AttributeError:
            response = h._conn.getresponse()

        # Add any cookie definitions to our list.
        for header in response.msg.getallmatchingheaders("Set-Cookie"):
            val = header.split(": ", 1)[1]
            cookie = val.split(";", 1)[0]
            self.cookies.append(cookie)

        if response.status != 200:
            raise xmlrpclib.ProtocolError(host + handler, response.status,
                                          response.reason,
                                          response.msg.headers)

        payload = response.read()
        parser, unmarshaller = self.getparser()
        parser.feed(payload)
        parser.close()

        return unmarshaller.close()


class BugzillaTransportMixin(CookieTransportMixin):

    LOGIN = 'Bugzilla_login'
    LOGIN_COOKIE = 'Bugzilla_logincookie'

    def remove_bugzilla_cookies(self):
        self.cookies = [x for x in self.cookies
                        if not x.startswith('%s=' % self.LOGIN) and
                        not x.startswith('%s=' % self.LOGIN_COOKIE)]

    def set_bugzilla_cookies(self, login, login_cookie):
        self.remove_bugzilla_cookies()
        self.cookies.append('%s=%s' % (self.LOGIN, login))
        self.cookies.append('%s=%s' % (self.LOGIN_COOKIE, login_cookie))

    def bugzilla_cookies(self):
        login = ''
        login_cookie = ''

        for c in self.cookies:
            name, _, val = c.partition('=')

            if name == self.LOGIN:
                login = val
            elif name == self.LOGIN_COOKIE:
                login_cookie = val

        return (login, login_cookie)


class BugzillaTransport(BugzillaTransportMixin, xmlrpclib.Transport):
    pass


class BugzillaSafeTransport(BugzillaTransportMixin, xmlrpclib.SafeTransport):
    pass


def bugzilla_transport(uri):
    """Return an appropriate Transport for the URI.

    If the URI type is https, return a CookieSafeTransport.
    If the type is http, return a CookieTransport.
    """
    if urlparse.urlparse(uri, "http")[0] == "https":
        return BugzillaSafeTransport()

    return BugzillaTransport()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from rbbz.auth import BugzillaBackend
from reviewboard.accounts.backends import get_enabled_auth_backends


class BugzillaCookieAuthMiddleware(object):
    """Set Bugzilla login cookies from auth backend."""

    def process_request(self, request):
        if BugzillaBackend not in [x.__class__ for
                                   x in get_enabled_auth_backends()]:
            return

        request.user.bzlogin = request.session.get('Bugzilla_login')
        request.user.bzcookie = request.session.get('Bugzilla_logincookie')

    def process_response(self, request, response):
        if BugzillaBackend not in [x.__class__ for
                                   x in get_enabled_auth_backends()]:
            return response

        if not hasattr(request, 'user'):
            return response

        if not request.user.is_authenticated():
            for key in ('Bugzilla_login', 'Bugzilla_logincookie'):
                try:
                    del request.session[key]
                except KeyError:
                    pass

            return response

        try:
            bzlogin = getattr(request.user, 'bzlogin')
            bzcookie = getattr(request.user, 'bzcookie')
        except AttributeError:
            return response

        if not bzlogin or not bzcookie:
            return response

        request.session['Bugzilla_login'] = bzlogin
        request.session['Bugzilla_logincookie'] = bzcookie
        return response



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import HTMLParser

from reviewboard.diffviewer.diffutils import (get_diff_files,
                                              populate_diff_chunks)


def get_file_chunks_in_range_custom(context, filediff, interfilediff,
                                    first_line, num_lines):
    """
    A generator that yields chunks within a range of lines in the specified
    filediff/interfilediff.

    This function was mostly copied from the Review Board source code. The
    main modification is to always set syntax highlighting to False so that
    plain code is returned in the chunks instead of html.

    This is primarily intended for use with templates. It takes a
    RequestContext for looking up the user and for caching file lists,
    in order to improve performance and reduce lookup times for files that have
    already been fetched.

    Each returned chunk is a dictionary with the following fields:

      ============= ========================================================
      Variable      Description
      ============= ========================================================
      ``change``    The change type ("equal", "replace", "insert", "delete")
      ``numlines``  The number of lines in the chunk.
      ``lines``     The list of lines in the chunk.
      ``meta``      A dictionary containing metadata on the chunk
      ============= ========================================================


    Each line in the list of lines is an array with the following data:

      ======== =============================================================
      Index    Description
      ======== =============================================================
      0        Virtual line number (union of the original and patched files)
      1        Real line number in the original file
      2        HTML markup of the original file
      3        Changed regions of the original line (for "replace" chunks)
      4        Real line number in the patched file
      5        HTML markup of the patched file
      6        Changed regions of the patched line (for "replace" chunks)
      7        True if line consists of only whitespace changes
      ======== =============================================================
    """
    def find_header(headers):
        for header in reversed(headers):
            if header[0] < first_line:
                return {
                    'line': header[0],
                    'text': header[1],
                }

    interdiffset = None

    key = "_diff_files_%s_%s" % (filediff.diffset.id, filediff.id)

    if interfilediff:
        key += "_%s" % (interfilediff.id)
        interdiffset = interfilediff.diffset

    if key in context:
        files = context[key]
    else:
        assert 'user' in context

        request = context.get('request', None)
        files = get_diff_files(filediff.diffset, filediff, interdiffset,
                               request=request)
        populate_diff_chunks(files, False,
                             request=request)
        context[key] = files

    if not files:
        raise StopIteration

    assert len(files) == 1
    last_header = [None, None]

    for chunk in files[0]['chunks']:
        if ('headers' in chunk['meta'] and
                (chunk['meta']['headers'][0] or chunk['meta']['headers'][1])):
            last_header = chunk['meta']['headers']

        lines = chunk['lines']

        if lines[-1][0] >= first_line >= lines[0][0]:
            start_index = first_line - lines[0][0]

            if first_line + num_lines <= lines[-1][0]:
                last_index = start_index + num_lines
            else:
                last_index = len(lines)

            new_chunk = {
                'lines': chunk['lines'][start_index:last_index],
                'numlines': last_index - start_index,
                'change': chunk['change'],
                'meta': chunk.get('meta', {}),
            }

            if 'left_headers' in chunk['meta']:
                left_header = find_header(chunk['meta']['left_headers'])
                right_header = find_header(chunk['meta']['right_headers'])
                del new_chunk['meta']['left_headers']
                del new_chunk['meta']['right_headers']

                if left_header or right_header:
                    header = (left_header, right_header)
                else:
                    header = last_header

                new_chunk['meta']['headers'] = header

            yield new_chunk

            first_line += new_chunk['numlines']
            num_lines -= new_chunk['numlines']

            assert num_lines >= 0
            if num_lines == 0:
                break


def render_equal_chunk(chunk, parser):
    indents = chunk['meta'].get('indentation_changes', {})
    lines = [] # Rendered lines.
    indent_count = 0 # How many indentation lines have we seen in a row.

    for line in chunk['lines']:
        indent_key = "%s-%s" % (line[1], line[4])

        if indent_key not in indents:
            indent_count = 0
            lines.append(">  %s" % parser.unescape(line[5]))
            continue

        # Unpack the indent data., is_indent indicates if the change added
        # more indentation (True added, False removed), num_chars is the
        # raw count of spaces and tabs in the indentation change, and
        # normalized_len is the length of the indentation change where
        # tab characters are treated as 8 spaces.
        #
        # e.g. if the following characters were added:
        #    "   \t  "
        # it would result in the following data:
        #     (True, 6, 10)
        # and it would appear in the line as:
        #    <span class="indent">&gt;&gt;&gt;&gt;---|&gt;&gt;</span>
        is_indent, num_chars, normalized_len = indents[indent_key]

        chars = None
        remainder = None
        original_len = len(line[2])
        patched_len = len(line[5])

        # Grab the characters we need to parse.
        if is_indent:
            chars = line[5][21:(patched_len - original_len)]
            remainder = line[2]
        else:
            remainder = line[5]
            chars = line[2][21:(original_len - patched_len)]

        index = 0
        replace_chars = []
        current_width = 0

        # Parse the indentation characters
        while index < len(chars):
            if chars[index] == "&" and chars[index + 1] == "g": # "&gt;".
                # Space character
                replace_chars.append(" ");
                current_width += 1
                index += 4

            elif chars[index] == "&" and chars[index + 1] == "m": # "&mdash;".
                # One of the spaces we translated before
                # was actually part of this tab we've
                # found
                replace_chars.pop()

                # find the end of the tab.
                while chars[index] != "|":
                    current_width += 1
                    index += 7

                replace_chars.append("\t")
                current_width += 1
                index += 1

            elif chars[index] == "|":
                # We've hit an ambiguous case. The previous character might
                # be a space or might be part of this tab. For now, lets just
                # be stupid and assume it's part of the tab (mixed indentation
                # is evil anyways).
                replace_chars.pop()
                replace_chars.append("\t")
                current_width += 1
                index += 1

            else:
                # We really shouldn't hit this case... if we did, something
                # went terribly wrong. lets just fill the rest of the
                # characters with spaces and bail
                replace_chars.append(" " * (normalized_len - current_width))
                index = len(chars)

        add_line = None
        remove_line = None
        remainder = parser.unescape(remainder)

        if is_indent:
            add_line = "> +%s%s" % ("".join(replace_chars), remainder)
            remove_line = "> -%s" % remainder
        else:
            add_line = "> +%s" % remainder
            remove_line = "> -%s%s" % ("".join(replace_chars), remainder)

        lines.insert(len(lines) - indent_count, remove_line)
        lines.append(add_line)

    return lines

def render_comment_plain(comment, context):
    parser = HTMLParser.HTMLParser()
    chunks = list(get_file_chunks_in_range_custom(
        context,
        comment.filediff,
        comment.interfilediff,
        comment.first_line,
        comment.num_lines))

    lines = [
        "::: %s" % comment.filediff.dest_file_display,
    ]

    if comment.interfilediff:
        lines.append(
            "(Diff revisions %s - %s)" % (
                comment.filediff.diffset.revision,
                comment.interfilediff.diffset.revision))
    else:
        lines.append(
            "(Diff revision %s)" % comment.filediff.diffset.revision)

    for chunk in chunks:
        if chunk['change'] == "equal":
            lines.extend(render_equal_chunk(chunk, parser))
        elif chunk['change'] == "insert":
            for line in chunk['lines']:
                lines.append("> +%s" % parser.unescape(line[5]))
        elif chunk['change'] == "delete":
            for line in chunk['lines']:
                lines.append("> -%s" % parser.unescape(line[2]))
        elif chunk['change'] == "replace":
            for line in chunk['lines']:
                lines.append("> -%s" % parser.unescape(line[2]))
            for line in chunk['lines']:
                lines.append("> +%s" % parser.unescape(line[5]))

    lines.append("")

    comments = []
    c = comment
    depth = 0

    while True:
        if depth:
            prefix = '%s ' % ('>' * depth,)
        else:
            prefix = ''

        comments.append("%s%s" % (prefix, c))

        if c.reply_to:
            c = c.reply_to
            depth += 1
        else:
            break

    comments.reverse()

    lines.extend(comments)

    return "\n".join(lines)


def build_plaintext_review(review, context):
    """Create a plaintext patch style representation of a review"""
    review_text = []

    if review.body_top:
        review_text.append(review.body_top)

    for comment in review.comments.all():
        review_text.append(render_comment_plain(comment, context))

    if review.body_bottom:
        review_text.append(review.body_bottom)

    return "\n\n".join(review_text)

